# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_W='Unable to parse "input" JSON document: '
_V='malformed-message'
_U='application/yang-data+json'
_T=':tenants/tenant=[^ ]*'
_S=':tenants/tenant=[^/]*/'
_R='" prefix.'
_Q='Top node names must begin with the "'
_P='application'
_O=False
_N='name'
_M=True
_L=':" prefix.'
_K='Non-root data_paths must begin with the "/'
_J=':tenants/tenant/0/'
_I=':tenant'
_H='+'
_G='invalid-value'
_F='text/plain'
_E='protocol'
_D=None
_C=':tenants/tenant='
_B=':'
_A='/'
import re,json,datetime,basicauth
from aiohttp import web
from passlib.hash import sha256_crypt
from .  import dal
from .  import utils
from .rcsvr import RestconfServer
from .handler import RouteHandler
class TenantViewHandler(RouteHandler):
	supported_media_types=_U,
	def __init__(A,native,yl_cb_func):B=native;A.native=B;A.yl_cb_func=yl_cb_func;assert B.dal.app_ns.endswith('-x');A.facade_ns=B.dal.app_ns.replace('-x','-1')
	async def _check_auth(C,request,data_path):
		K='access-denied';J='comment';I='failure';H='outcome';E=request;A={};A['timestamp']=datetime.datetime.utcnow();A['source-ip']=E.remote;A['source-proxies']=list(E.forwarded);A['host']=E.host;A['method']=E.method;A['path']=E.path;L=E.headers.get('AUTHORIZATION')
		if L is _D:A[H]=I;A[J]='No authorization specified in the HTTP header.';await C.native._insert_audit_log_entry(_D,A);B=web.Response(status=401);D=utils.gen_rc_errors(_E,K);B.text=json.dumps(D,indent=2);return B
		G,N=basicauth.decode(L);F=_D
		try:F=await C.native.dal.get_tenant_name_for_admin(G)
		except dal.NodeNotFound as Q:A[H]=I;A[J]='Unknown admin: '+G;await C.native._insert_audit_log_entry(_D,A);B=web.Response(status=401);D=utils.gen_rc_errors(_E,K);B.text=json.dumps(D,indent=2);return B
		if F==_D:A[H]=I;A[J]='Host-level admins cannot use tenant interface ('+G+').';await C.native._insert_audit_log_entry(_D,A);B=web.Response(status=401);D=utils.gen_rc_errors(_E,K);B.text=json.dumps(D,indent=2);return B
		O=_A+C.native.dal.app_ns+_C+F+'/admin-accounts/admin-account='+G+'/password';P=await C.native.dal.handle_get_config_request(O);M=P[C.native.dal.app_ns+':password'];assert M.startswith('$5$')
		if not sha256_crypt.verify(N,M):A[H]=I;A[J]='Password mismatch for admin '+G;await C.native._insert_audit_log_entry(F,A);B=web.Response(status=401);D=utils.gen_rc_errors(_E,K);B.text=json.dumps(D,indent=2);return B
		A[H]='success';await C.native._insert_audit_log_entry(F,A);return F
	async def handle_get_restconf_root(E,request):
		F=request;H=_A;C=await E._check_auth(F,H)
		if type(C)is web.Response:A=C;return A
		else:I=C
		B,J=utils.check_http_headers(F,E.supported_media_types,accept_required=_M)
		if type(B)is web.Response:A=B;return A
		else:assert type(B)==str;D=B;assert D!=_F;G=utils.Encoding[D.rsplit(_H,1)[1]]
		A=web.Response(status=200);A.content_type=D
		if G==utils.Encoding.json:A.text='{\n    "ietf-restconf:restconf" : {\n        "data" : {},\n        "operations" : {},\n        "yang-library-version" : "2019-01-04"\n    }\n}\n'
		else:assert G==utils.Encoding.xml;A.text='<restconf xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">\n    <data/>\n    <operations/>\n    <yang-library-version>2016-06-21</yang-library-version>\n</restconf>\n'
		return A
	async def handle_get_yang_library_version(E,request):
		F=request;H=_A;C=await E._check_auth(F,H)
		if type(C)is web.Response:A=C;return A
		else:I=C
		B,J=utils.check_http_headers(F,E.supported_media_types,accept_required=_M)
		if type(B)is web.Response:A=B;return A
		else:assert type(B)==str;D=B;assert D!=_F;G=utils.Encoding[D.rsplit(_H,1)[1]]
		A=web.Response(status=200);A.content_type=D
		if G==utils.Encoding.json:A.text='{\n  "ietf-restconf:yang-library-version" : "2019-01-04"\n}'
		else:assert G==utils.Encoding.xml;A.text='<yang-library-version xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">2019-01-04</yang-library-version>'
		return A
	async def handle_get_opstate_request(A,request):
		F=request;C,M=utils.parse_raw_path(F._message.path[RestconfServer.len_prefix_operational:]);G=await A._check_auth(F,C)
		if type(G)is web.Response:B=G;return B
		else:H=G
		E,W=utils.check_http_headers(F,A.supported_media_types,accept_required=_M)
		if type(E)is web.Response:B=E;return B
		else:assert type(E)==str;N=E;assert N!=_F;X=utils.Encoding[N.rsplit(_H,1)[1]]
		if C=='/ietf-yang-library:yang-library':B=web.Response(status=200);B.content_type=_U;B.text=A.yl_cb_func();return B
		assert C==_A or C.startswith(_A+A.facade_ns+_B)
		if C==_A:O=_A+A.native.dal.app_ns+_C+H
		else:
			if not C.startswith(_A+A.facade_ns+_B):B=web.Response(status=400);S=_K+A.facade_ns+_L;T=utils.gen_rc_errors(_E,_G,error_message=S);B.text=json.dumps(T,indent=2);return B
			Y,P=C.split(_B,1);assert P!=_D;O=_A+A.native.dal.app_ns+_C+H+_A+P
		Q=dict()
		for R in M.keys():Q[R]=re.sub(_A+A.facade_ns+_B,_A+A.native.dal.app_ns+_C+H+_A,M[R])
		I,D=await A.native.handle_get_opstate_request_lower_half(O,Q)
		if D!=_D:
			assert I.status==200;J={}
			if C==_A:
				for K in D[A.native.dal.app_ns+_I][0].keys():
					if K==_N:continue
					J[A.facade_ns+_B+K]=D[A.native.dal.app_ns+_I][0][K]
			else:L=next(iter(D));assert L.count(_B)==1;U,V=L.split(_B);assert U==A.native.dal.app_ns+'';assert type(D)==dict;assert len(D)==1;J[A.facade_ns+_B+V]=D[L]
			I.text=json.dumps(J,indent=2)
		return I
	async def handle_get_config_request(A,request):
		F=request;B,M=utils.parse_raw_path(F._message.path[RestconfServer.len_prefix_running:]);G=await A._check_auth(F,B)
		if type(G)is web.Response:D=G;return D
		else:H=G
		E,W=utils.check_http_headers(F,A.supported_media_types,accept_required=_M)
		if type(E)is web.Response:D=E;return D
		else:assert type(E)==str;N=E;assert N!=_F;X=utils.Encoding[N.rsplit(_H,1)[1]]
		assert B==_A or B.startswith(_A+A.facade_ns+_B)
		if B==_A:O=_A+A.native.dal.app_ns+_C+H
		else:
			if not B.startswith(_A+A.facade_ns+_B):D=web.Response(status=400);S=_K+A.facade_ns+_L;T=utils.gen_rc_errors(_E,_G,error_message=S);D.text=json.dumps(T,indent=2);return D
			Y,P=B.split(_B,1);assert P!=_D;O=_A+A.native.dal.app_ns+_C+H+_A+P
		Q=dict()
		for R in M.keys():Q[R]=re.sub(_A+A.facade_ns+_B,_A+A.native.dal.app_ns+_C+H+_A,M[R])
		I,C=await A.native.handle_get_config_request_lower_half(O,Q)
		if C!=_D:
			assert I.status==200;J={}
			if B==_A:
				for K in C[A.native.dal.app_ns+_I][0].keys():
					if K==_N:continue
					J[A.facade_ns+_B+K]=C[A.native.dal.app_ns+_I][0][K]
			else:L=next(iter(C));assert L.count(_B)==1;U,V=L.split(_B);assert U==A.native.dal.app_ns+'';assert type(C)==dict;assert len(C)==1;J[A.facade_ns+_B+V]=C[L]
			I.text=json.dumps(J,indent=2)
		return I
	async def handle_post_config_request(B,request):
		D=request;E,K=utils.parse_raw_path(D._message.path[RestconfServer.len_prefix_running:]);H=await B._check_auth(D,E)
		if type(H)is web.Response:A=H;return A
		else:I=H
		F,V=utils.check_http_headers(D,B.supported_media_types,accept_required=_O)
		if type(F)is web.Response:A=F;return A
		else:assert type(F)==str;L=F;assert L!=_F;W=utils.Encoding[L.rsplit(_H,1)[1]]
		if E==_A:M=_A+B.native.dal.app_ns+_C+I
		else:
			if not E.startswith(_A+B.facade_ns+_B):A=web.Response(status=400);Q=_K+B.facade_ns+_L;C=utils.gen_rc_errors(_E,_G,error_message=Q);A.text=json.dumps(C,indent=2);return A
			X,N=E.split(_B,1);assert N!=_D;M=_A+B.native.dal.app_ns+_C+I+_A+N
		O=dict()
		for P in K.keys():O[P]=re.sub(_A+B.facade_ns+_B,_A+B.native.dal.app_ns+_C+I+_A,K[P])
		try:G=await D.json()
		except json.decoder.JSONDecodeError as R:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_V,error_message=_W+str(R));A.text=json.dumps(C,indent=2);return A
		assert type(G)==dict;assert len(G)==1;J=next(iter(G));assert J.count(_B)==1;S,T=J.split(_B)
		if S!=B.facade_ns:A=web.Response(status=400);C=utils.gen_rc_errors(_P,_G,error_message=_Q+B.facade_ns+_R);A.text=json.dumps(C,indent=2);return A
		U={B.native.dal.app_ns+_B+T:G[J]};A=await B.native.handle_post_config_request_lower_half(M,O,U)
		if A.status!=201:
			if _A+B.native.dal.app_ns+_J in A.text:A.text=A.text.replace(_A+B.native.dal.app_ns+_J,_A+B.facade_ns+_B)
			elif _A+B.native.dal.app_ns+_C in A.text:A.text=re.sub(_A+B.native.dal.app_ns+_S,_A+B.facade_ns+_B,A.text);A.text=re.sub(_A+B.native.dal.app_ns+_T,_A+B.facade_ns+_B,A.text)
		return A
	async def handle_put_config_request(B,request):
		F=request;E,M=utils.parse_raw_path(F._message.path[RestconfServer.len_prefix_running:]);I=await B._check_auth(F,E)
		if type(I)is web.Response:A=I;return A
		else:G=I
		H,Y=utils.check_http_headers(F,B.supported_media_types,accept_required=_O)
		if type(H)is web.Response:A=H;return A
		else:assert type(H)==str;N=H;assert N!=_F;Z=utils.Encoding[N.rsplit(_H,1)[1]]
		if E==_A:O=_A+B.native.dal.app_ns+_C+G
		else:
			if not E.startswith(_A+B.facade_ns+_B):A=web.Response(status=400);S=_K+B.facade_ns+_L;C=utils.gen_rc_errors(_E,_G,error_message=S);A.text=json.dumps(C,indent=2);return A
			a,P=E.split(_B,1);assert P!=_D;O=_A+B.native.dal.app_ns+_C+G+_A+P
		Q=dict()
		for R in M.keys():Q[R]=re.sub(_A+B.facade_ns+_B,_A+B.native.dal.app_ns+_C+G+_A,M[R])
		try:D=await F.json()
		except json.decoder.JSONDecodeError as T:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_V,error_message=_W+str(T));A.text=json.dumps(C,indent=2);return A
		if E==_A:
			J={B.native.dal.app_ns+_I:[{_N:G}]}
			for K in D.keys():
				assert K.count(_B)==1;U,V=K.split(_B)
				if U!=B.facade_ns:A=web.Response(status=400);C=utils.gen_rc_errors(_P,_G,error_message=_Q+B.facade_ns+_R);A.text=json.dumps(C,indent=2);return A
				J[B.native.dal.app_ns+_I][0][V]=D[K]
		else:
			assert type(D)==dict;assert len(D)==1;L=next(iter(D));assert L.count(_B)==1;W,X=L.split(_B)
			if W!=B.facade_ns:A=web.Response(status=400);C=utils.gen_rc_errors(_P,_G,error_message=_Q+B.facade_ns+_R);A.text=json.dumps(C,indent=2);return A
			J={B.native.dal.app_ns+_B+X:D[L]}
		A=await B.native.handle_put_config_request_lower_half(O,Q,J)
		if A.status!=201 and A.status!=204:
			if _A+B.native.dal.app_ns+_J in A.text:A.text=A.text.replace(_A+B.native.dal.app_ns+_J,_A+B.facade_ns+_B)
			elif _A+B.native.dal.app_ns+_C in A.text:A.text=re.sub(_A+B.native.dal.app_ns+_S,_A+B.facade_ns+_B,A.text);A.text=re.sub(_A+B.native.dal.app_ns+_T,_A+B.facade_ns+_B,A.text)
		return A
	async def handle_delete_config_request(B,request):
		F=request;C,N=utils.parse_raw_path(F._message.path[RestconfServer.len_prefix_running:]);G=await B._check_auth(F,C)
		if type(G)is web.Response:A=G;return A
		else:H=G
		D,O=utils.check_http_headers(F,B.supported_media_types,accept_required=_O)
		if type(D)is web.Response:A=D;return A
		else:
			assert type(D)==str;E=D
			if E==_F:L=_D
			else:L=utils.Encoding[E.rsplit(_H,1)[1]]
		if C==_A:I=_A+B.native.dal.app_ns+_C+H
		else:
			if not C.startswith(_A+B.facade_ns+_B):
				J=_K+B.facade_ns+_L;A=web.Response(status=400);A.content_type=E
				if E==_F:A.text=J
				else:M=utils.gen_rc_errors(_E,_G,error_message=J);A.text=json.dumps(M,indent=2)
				return A
			P,K=C.split(_B,1);assert K!=_D;I=_A+B.native.dal.app_ns+_C+H+_A+K
		A=await B.native.handle_delete_config_request_lower_half(I)
		if A.status!=204:
			if _A+B.native.dal.app_ns+_J in A.text:A.text=A.text.replace(_A+B.native.dal.app_ns+_J,_A+B.facade_ns+_B)
			elif _A+B.native.dal.app_ns+_C in A.text:A.text=re.sub(_A+B.native.dal.app_ns+_S,_A+B.facade_ns+_B,A.text);A.text=re.sub(_A+B.native.dal.app_ns+_T,_A+B.facade_ns+_B,A.text)
		return A
	async def handle_action_request(A,request):0
	async def handle_rpc_request(A,request):0


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

_F='tested?'
_E='wn-sztpd-0:device'
_D='device'
_C='activation-code'
_B='/'
_A=None
import gc,tracemalloc,os,re,json,base64,signal,asyncio,datetime,functools,pkg_resources
from passlib.hash import sha256_crypt
from .yangcore import utils
from .yangcore.rcsvr import RestconfServer
from .yangcore.tenant import TenantViewHandler
from .yangcore.native import NativeViewHandler,Period,TimeUnit
from .yangcore.dal import DataAccessLayer,CreateCallbackFailed,CreateOrChangeCallbackFailed,ChangeCallbackFailed,AuthenticationFailed
from .rfc8572 import RFC8572ViewHandler
from .  import yl
from pyasn1.codec.der.decoder import decode as decode_der
from pyasn1.error import PyAsn1Error
from pyasn1_modules import rfc5652
from cryptography import x509
from cryptography.hazmat.primitives import serialization
loop=_A
sig=_A
def signal_handler(name):global loop;global sig;sig=name;loop.stop()
def run(db_url,cacert_param=_A,cert_param=_A,key_param=_A):
	b=':tenants/tenant/bootstrap-servers/bootstrap-server/trust-anchor';a=':bootstrap-servers/bootstrap-server/trust-anchor';Z=':transport';Y='SIGHUP';X='Yes';S='use-for';R='1';O=db_url;N='x';L=key_param;K=cert_param;J=cacert_param;global loop;global sig;A=_A;B=_A;T=False
	if J is not _A and O.startswith('sqlite:'):print('The "sqlite" dialect does not support the "cacert" parameter.');return 1
	if(K or L)and not J:print('The "cacert" parameter must be specified whenever the "key" and "cert" parameters are specified.');return 1
	if(K is _A)!=(L is _A):print('The "key" and "cert" parameters must be specified together.');return 1
	try:A=DataAccessLayer(O,J,K,L)
	except (SyntaxError,AssertionError,AuthenticationFailed)as H:print(str(H));return 1
	except NotImplementedError as H:T=True
	else:B=A.opaque()
	if T==True:
		D=os.environ.get('SZTPD_ACCEPT_CONTRACT')
		if D==_A:
			print('');c=pkg_resources.resource_filename('sztpd','LICENSE.txt');U=open(c,'r');print(U.read());U.close();print('First time initialization.  Please accept the license terms.');print('');print('By entering "Yes" below, you agree to be bound to the terms and conditions contained on this screen with Watsen Networks.');print('');d=input('Please enter "Yes" or "No": ')
			if d!=X:print('');print('Thank you for your consideration.');print('');return 1
		elif D!=X:print('');print('The "SZTPD_ACCEPT_CONTRACT" environment variable is set to a value other than "Yes".  Please correct the value and try again.');print('');return 1
		D=os.environ.get('SZTPD_INIT_MODE')
		if D==_A:
			print('');print('Modes:');print('  1 - single-tenant');print('  x - multi-tenant');print('');B=input('Please select mode: ')
			if B not in[R,N]:print('Unknown mode selection.  Please try again.');return 1
			print('');print("Running SZTPD in mode '"+B+"'. (No more output expected)");print('')
		elif D not in[R,N]:print('The "SZTPD_INIT_MODE" environment variable is set to an unknown value.  Must be \'1\' or \'x\'.');return 1
		else:B=D
		D=os.environ.get('SZTPD_INIT_PORT')
		if D!=_A:
			try:V=int(D)
			except ValueError as H:print('Invalid "SZTPD_INIT_PORT" value ('+D+').');return 1
			if V<=0 or V>2**16-1:print('The "SZTPD_INIT_PORT" value ('+D+') is out of range [1..65535].');return 1
		try:A=DataAccessLayer(O,J,K,L,json.loads(getattr(yl,'nbi_'+B)()),'wn-sztpd-'+B,B)
		except Exception as H:raise H;return 1
	assert B!=_A;assert A!=_A;tracemalloc.start(25);loop=asyncio.get_event_loop();loop.add_signal_handler(signal.SIGHUP,functools.partial(signal_handler,name=Y));loop.add_signal_handler(signal.SIGTERM,functools.partial(signal_handler,name='SIGTERM'));loop.add_signal_handler(signal.SIGINT,functools.partial(signal_handler,name='SIGINT'));loop.add_signal_handler(signal.SIGQUIT,functools.partial(signal_handler,name='SIGQUIT'))
	while sig is _A:
		M=[];F=A.handle_get_config_request(_B+A.app_ns+Z);P=loop.run_until_complete(F)
		for E in P[A.app_ns+Z]['listen']['endpoint']:
			if E[S]=='native-interface':
				C=NativeViewHandler(A,B,loop)
				if B=='0':G=_B+A.app_ns+':device'
				elif B==R:G=_B+A.app_ns+':devices/device'
				elif B==N:G=_B+A.app_ns+':tenants/tenant/devices/device'
				C.register_create_callback(G,_handle_device_created);e=G+'/activation-code';C.register_change_callback(e,_handle_device_act_code_changed);C.register_subtree_change_callback(G,_handle_device_subtree_changed);C.register_somehow_change_callback(G,_handle_device_somehow_changed);C.register_delete_callback(G,_handle_device_deleted);C.register_periodic_callback(Period(24,TimeUnit.Hours),datetime.datetime(2000,1,1,0),_check_expirations)
				if B!=N:C.register_create_callback(_B+A.app_ns+a,_handle_bss_trust_anchor_cert_created_or_changed);C.register_change_callback(_B+A.app_ns+a,_handle_bss_trust_anchor_cert_created_or_changed)
				else:C.register_create_callback(_B+A.app_ns+b,_handle_bss_trust_anchor_cert_created_or_changed);C.register_change_callback(_B+A.app_ns+b,_handle_bss_trust_anchor_cert_created_or_changed)
				Q=RestconfServer(loop,A,E,C)
			elif E[S]=='tenant-interface':
				def f():return getattr(yl,'nbi_x_tenant')()
				g=TenantViewHandler(C,f);Q=RestconfServer(loop,A,E,g)
			else:assert E[S]=='rfc8572-interface';W=json.loads(getattr(yl,'sbi_rfc8572')());h=RFC8572ViewHandler(A,B,W,C);Q=RestconfServer(loop,A,E,h,W)
			M.append(Q);del E;E=_A
		del P;P=_A;loop.run_forever()
		for I in M:F=I.app.shutdown();loop.run_until_complete(F);F=I.runner.cleanup();loop.run_until_complete(F);F=I.app.cleanup();loop.run_until_complete(F);del I;I=_A
		del M;M=_A
		if sig==Y:sig=_A
	loop.close();del A;return 0
async def _handle_device_created_post_sweep(watched_node_path,conn,opaque):
	g=':dynamic-callout';f='webhooks';e='verification-result';d='failure';c='tenant';b='function';a='functions';Z='plugin';Y='callback';X='ownership-authorization';O='verification-results';N='dynamic-callout';M='device-type';L='row_id';K='=[^/]*';I=watched_node_path;H='wn-sztpd-rpcs:output';B=conn;A=opaque;C=A.dal._get_row_data_for_list_path(I,B);D=re.sub(K,'',I);h=A.dal._get_jsob_for_row_id_in_table(D,C[L],B);P=_B+A.dal.app_ns+':device-types/device-type='+h[_D][M];C=A.dal._get_row_data_for_list_path(P,B);D=re.sub(K,'',P);Q=A.dal._get_jsob_for_row_id_in_table(D,C[L],B)
	if X in Q[M]:
		R=_B+A.dal.app_ns+':dynamic-callouts/dynamic-callout='+Q[M][X][N]['reference'];C=A.dal._get_row_data_for_list_path(R,B);D=re.sub(K,'',R);E=A.dal._get_jsob_for_row_id_in_table(D,C[L],B)
		if Y in E[N]:
			F=E[N][Y];assert F[Z]in A.plugins;S=A.plugins[F[Z]];assert a in S;T=S[a];assert F[b]in T;i=T[F[b]];J=I.split(_B)
			if J[2]==c:U=J[1].split('=')[1]
			else:U='not-applicable'
			V=J[-1].split('=')[1];j={'wn-sztpd-rpcs:input':{c:U,'serial-number':[V]}};G=i(j);W=d
			if H in G:
				if O in G[H]:
					if e in G[H][O]:W=G[H][O][e][0]['result']
			if W==d:raise CreateCallbackFailed('Unable to verify ownership for device: '+V)
		else:assert f in E[A.dal.app_ns+g][0];k=E[A.dal.app_ns+g][0][f];raise NotImplementedError('webhooks for ownership verification not implemented yet')
async def _handle_device_created(watched_node_path,jsob,jsob_data_path,nvh):
	C=nvh;B=jsob;assert type(B)==dict
	if jsob_data_path==_B:assert _E in B;A=B[_E]
	else:assert _D in B;A=B[_D]
	if C.dal.post_dal_callbacks is _A:C.dal.post_dal_callbacks=[]
	C.dal.post_dal_callbacks.append((_handle_device_created_post_sweep,watched_node_path,C));A['lifecycle-statistics']={'nbi-access-stats':{'created':datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ'),'num-times-modified':0},'sbi-access-stats':{'num-times-accessed':0}};A['bootstrapping-log']={'log-entry':[]}
	if _C in A and A[_C].startswith('$0$'):A[_C]=sha256_crypt.using(rounds=1000).hash(A[_C][3:])
async def _handle_device_act_code_changed(watched_node_path,jsob,jsob_data_path,nvh):
	A=jsob;assert type(A)==dict
	if jsob_data_path==_B:assert _E in A;B=A[_E]
	else:assert _D in A;B=A[_D]
	if _C in B and B[_C].startswith('$0$'):B[_C]=sha256_crypt.using(rounds=1000).hash(B[_C][3:])
async def _handle_device_subtree_changed(watched_node_path,jsob,jsob_data_path,nvh):raise NotImplementedError(_F)
async def _handle_device_somehow_changed(watched_node_path,jsob,jsob_data_path,nvh):raise NotImplementedError(_F)
async def _handle_device_deleted(data_path,nvh):0
async def _handle_bss_trust_anchor_cert_created_or_changed(watched_node_path,jsob,jsob_data_path,obj):
	G='": ';B=watched_node_path;H=jsob['bootstrap-server']['trust-anchor'];I=base64.b64decode(H)
	try:J,O=decode_der(I,asn1Spec=rfc5652.ContentInfo())
	except PyAsn1Error as K:raise CreateOrChangeCallbackFailed('Parsing trust anchor certificate CMS structure failed for '+B+' ('+str(K)+')')
	L=utils.degenerate_cms_obj_to_ders(J);A=[]
	for M in L:N=x509.load_der_x509_certificate(M);A.append(N)
	D=[B for B in A if B.subject==B.issuer]
	if len(D)==0:raise CreateOrChangeCallbackFailed('Trust anchor certificates must encode a root (self-signed) certificate: '+B)
	if len(D)>1:raise CreateOrChangeCallbackFailed('Trust anchor certificates must encode no more than one root (self-signed) certificate ('+str(len(D))+' found): '+B)
	F=D[0];A.remove(F);C=F
	while len(A):
		E=[B for B in A if B.issuer==C.subject]
		if len(E)==0:raise CreateOrChangeCallbackFailed('Trust anchor certificates must not encode superfluous certificates.  CMS encodes additional certs not issued by the certificate "'+str(C.subject)+G+B)
		if len(E)>1:raise CreateOrChangeCallbackFailed('Trust anchor certificates must encode a single chain of certificates.  Found '+str(len(E))+' certificates issued by "'+str(C.subject)+G+B)
		C=E[0];A.remove(C)
def _check_expirations(nvh):0


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_C='/ds/ietf-datastores:operational'
_B='/ds/ietf-datastores:running'
_A=None
import os,ssl,json,base64,pyasn1,asyncio,yangson,datetime,tempfile,basicauth
from .  import utils
from aiohttp import web
from .handler import RouteHandler
from pyasn1.type import univ
from pyasn1_modules import rfc3447
from pyasn1_modules import rfc5280
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5915
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.codec.der.encoder import encode as der_encoder
async def set_server_header(request,response):response.headers['Server']='<redacted>'
class RestconfServer:
	root='/restconf';prefix_running=root+_B;prefix_operational=root+_C;prefix_operations=root+'/operations';len_prefix_running=len(prefix_running);len_prefix_operational=len(prefix_operational);len_prefix_operations=len(prefix_operations)
	def __init__(A,loop,dal,endpoint_config,view_handler,facade_yl=_A):
		w='client-certs';v='local-truststore-reference';u='ca-certs';t='client-authentication';s='\n-----END CERTIFICATE-----\n';r='-----BEGIN CERTIFICATE-----\n';q='cert-data';p='private-key-format';o=':keystore/asymmetric-keys/asymmetric-key=';n='reference';m='server-identity';l='local-port';k='http';V='ASCII';U=':asymmetric-key';T='tcp-server-parameters';M='certificate';L='tls-server-parameters';K='/ds/ietf-datastores:running{tail:.*}';J='/';G=dal;D='https';C=endpoint_config;B=view_handler;A.len_prefix_running=len(A.root+_B);A.len_prefix_operational=len(A.root+_C);A.loop=loop;A.dal=G;A.name=C['name'];A.view_handler=B;A.app=web.Application(client_max_size=1024*1024*32);A.app.on_response_prepare.append(set_server_header);A.app.router.add_get('/.well-known/host-meta',A.handle_get_host_meta);A.app.router.add_get(A.root,B.handle_get_restconf_root);A.app.router.add_get(A.root+J,B.handle_get_restconf_root);A.app.router.add_get(A.root+'/yang-library-version',B.handle_get_yang_library_version);A.app.router.add_get(A.root+'/ds/ietf-datastores:operational{tail:.*}',B.handle_get_opstate_request);A.app.router.add_get(A.root+K,B.handle_get_config_request);A.app.router.add_put(A.root+K,B.handle_put_config_request);A.app.router.add_post(A.root+K,B.handle_post_config_request);A.app.router.add_delete(A.root+K,B.handle_delete_config_request);A.app.router.add_post(A.root+'/ds/ietf-datastores:operational/{tail:.*}',B.handle_action_request);A.app.router.add_post(A.root+'/operations/{tail:.*}',B.handle_rpc_request)
		if k in C:F=k
		else:assert D in C;F=D
		A.local_address=C[F][T]['local-address'];A.local_port=os.environ.get('SZTPD_INIT_PORT',8080)
		if l in C[F][T]:A.local_port=C[F][T][l]
		E=_A
		if F==D:
			W=C[D][L][m][M][n]['asymmetric-key'];N=A.dal.handle_get_config_request(J+A.dal.app_ns+o+W);O=A.loop.run_until_complete(N);P=O[A.dal.app_ns+U][0]['cleartext-private-key'];X=base64.b64decode(P)
			if O[A.dal.app_ns+U][0][p]=='ietf-crypto-types:ec-private-key-format':Q,x=der_decoder(X,asn1Spec=rfc5915.ECPrivateKey());y=der_encoder(Q);Y=base64.b64encode(y).decode(V);assert P==Y;Z='-----BEGIN EC PRIVATE KEY-----\n'+Y+'\n-----END EC PRIVATE KEY-----\n'
			elif O[A.dal.app_ns+U][0][p]=='ietf-crypto-types:rsa-private-key-format':Q,x=der_decoder(X,asn1Spec=rfc3447.RSAPrivateKey());z=der_encoder(Q);a=base64.b64encode(z).decode(V);assert P==a;Z='-----BEGIN RSA PRIVATE KEY-----\n'+a+'\n-----END RSA PRIVATE KEY-----\n'
			else:raise NotImplementedError('this line can never be reached')
			A0=C[D][L][m][M][n][M];N=A.dal.handle_get_config_request(J+A.dal.app_ns+o+W+'/certificates/certificate='+A0);A1=A.loop.run_until_complete(N);A2=A1[A.dal.app_ns+':certificate'][0][q];A3=base64.b64decode(A2);A4,b=der_decoder(A3,asn1Spec=rfc5652.ContentInfo());A5=A4.getComponentByName('content');A6,b=der_decoder(A5,asn1Spec=rfc5652.SignedData());c=A6.getComponentByName('certificates');H=''
			for A7 in range(len(c)):
				d=c[A7][0]
				for e in d['tbsCertificate']['extensions']:
					if e['extnID']==rfc5280.id_ce_basicConstraints:A8,b=der_decoder(e['extnValue'],asn1Spec=rfc5280.BasicConstraints())
				A9=der_encoder(d);f=base64.b64encode(A9).decode(V)
				if A8['cA']==False:H=r+f+s+H
				else:H+=r+f+s
			E=ssl.create_default_context(ssl.Purpose.CLIENT_AUTH);E.verify_mode=ssl.CERT_OPTIONAL
			with tempfile.TemporaryDirectory()as g:
				h=g+'key.pem';i=g+'certs.pem'
				with open(h,'w')as AA:AA.write(Z)
				with open(i,'w')as AB:AB.write(H)
				E.load_cert_chain(i,h)
			if t in C[D][L]:
				I=C[D][L][t]
				def j(truststore_ref):
					C=G.handle_get_config_request(J+G.app_ns+':truststore/certificate-bags/certificate-bag='+truststore_ref);D=A.loop.run_until_complete(C);B=[]
					for E in D[G.app_ns+':certificate-bag'][0][M]:F=base64.b64decode(E[q]);H,I=der_decoder(F,asn1Spec=rfc5652.ContentInfo());assert not I;B+=utils.degenerate_cms_obj_to_ders(H)
					return B
				R=[]
				if u in I:S=I[u][v];R+=j(S)
				if w in I:S=I[w][v];R+=j(S)
				AC=utils.der_dict_to_multipart_pem({'CERTIFICATE':R});E.load_verify_locations(cadata=AC)
		if F==D:assert not E is _A
		else:assert E is _A
		A.runner=web.AppRunner(A.app);A.loop.run_until_complete(A.runner.setup());A.site=web.TCPSite(A.runner,host=A.local_address,port=A.local_port,ssl_context=E,reuse_port=True);A.loop.run_until_complete(A.site.start())
	async def handle_get_host_meta(B,request):A=web.Response();A.content_type='application/xrd+xml';A.text='<XRD xmlns="http://docs.oasis-open.org/ns/xri/xrd-1.0">\n  <Link rel="restconf" href="/restconf"/>\n</XRD>';return A